module bellek_yonetimi_8;

interface Hayvan {
    string şarkıSöyle();
}

class Kedi : Hayvan {
    string şarkıSöyle() {
        return "miyav";
    }
}

class Papağan : Hayvan {
    string[] sözler;

    this(string[] sözler) {
        this.sözler = sözler;
    }

    string şarkıSöyle() {
        /* std.algorithm.joiner, belirtilen aralıktaki
         * elemanları belirtilen ayraçla birleştirir. */
        return sözler.joiner(", ").to!string;
    }
}
T * hizalıAdres(T)(T * adayAdres) {
    import std.traits;

    static if (is (T == class)) {
        const hizalama = classInstanceAlignment!T;

    } else {
        const hizalama = T.alignof;
    }

    const sonuç = (cast(size_t)adayAdres + hizalama - 1)
                  / hizalama * hizalama;
    return cast(T*)sonuç;
}
void * hizalıAdres(T)(void * adayAdres) {
    return hizalıAdres(cast(T*)adayAdres);
}
import std.stdio;
import std.algorithm;
import std.conv;
import core.memory;

// ...

void main() {
    /* Bu bir Hayvan değişkeni dizisidir; Hayvan nesnesi
     * dizisi değildir. */
    Hayvan[] hayvanlar;

    /* On bin baytın bu örnekte yeterli olduğunu varsayalım.
     * Normalde nesnelerin buraya gerçekten sığacaklarının da
     * denetlenmesi gerekir. */
    const sığa = 10_000;
    void * boşYer = GC.calloc(sığa);

    /* İlk önce bir Kedi nesnesi yerleştireceğiz. */
    void * kediAdayAdresi = boşYer;
    void * kediAdresi = hizalıAdres!Kedi(kediAdayAdresi);
    writeln("Kedi adresi   : ", kediAdresi);

    /* Sınıflarda emplace()'e void[] verildiğinden adresten
     * dilim elde etmek gerekiyor. */
    size_t kediUzunluğu = __traits(classInstanceSize, Kedi);
    void[] kediYeri = kediAdresi[0..kediUzunluğu];

    /* Kedi'yi o yerde kuruyoruz ve döndürülen değişkeni
     * diziye ekliyoruz. */
    Kedi kedi = emplace!Kedi(kediYeri);
    hayvanlar ~= kedi;

    /* Papağan'ı Kedi nesnesinden sonraki ilk uygun adreste
     * kuracağız. */
    void * papağanAdayAdresi = kediAdresi + kediUzunluğu;
    void * papağanAdresi =
        hizalıAdres!Papağan(papağanAdayAdresi);
    writeln("Papağan adresi: ", papağanAdresi);

    size_t papağanUzunluğu =
        __traits(classInstanceSize, Papağan);
    void[] papağanYeri = papağanAdresi[0..papağanUzunluğu];

    Papağan papağan =
        emplace!Papağan(papağanYeri, [ "merrba", "aloo" ]);
    hayvanlar ~= papağan;

    /* Nesneleri kullanıyoruz. */
    foreach (hayvan; hayvanlar) {
        writeln(hayvan.şarkıSöyle());
    }
}
